;; Generated package description from tuareg.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "tuareg" "3.0.20.20250910.140516" "OCaml mode" '((emacs "26.3") (caml "4.8")) :commit "f0cb55f2177f6fc978d98d018910fe5b1890fe0c" :authors '(("Albert Cohen" . "Albert.Cohen@inria.fr") ("Sam Steingold" . "sds@gnu.org") ("Christophe Troestler" . "Christophe.Troestler@umons.ac.be") ("Till Varoquaux" . "till@pps.jussieu.fr") ("Sean McLaughlin" . "seanmcl@gmail.com") ("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '(("Christophe Troestler" . "Christophe.Troestler@umons.ac.be") ("Stefan Monnier" . "monnier@iro.umontreal.ca")) :keywords '("ocaml" "languages") :url "https://github.com/ocaml/tuareg")
